@@ -1,19 +1,17 @@
 """MESSENGER UVVS data class"""
 import numpy as np
-from scipy.optimize import minimize_scalar
 import pandas as pd
-import bokeh.plotting as bkp
-from bokeh.models import HoverTool, Whisker
-from bokeh.palettes import Set1
-from bokeh.io import export_png
+import copy
 from astropy import units as u
-from astropy.visualization import PercentileInterval
+
+from nexoclom import Input, Output, LOSResult
 from .database_setup import database_connect
-from nexoclom import Input
+from .plot_methods import plot_bokeh, plot_plotly, plot_fitted
+from mathMB import fit_model
 
 
 class InputError(Exception):
-    """Raised when a required parameter is not included in the inputfile."""
+    """Raised when a required parameter is not included."""
     def __init__(self, expression, message):
         self.expression = expression
         self.message = message
@@ -36,7 +34,11 @@ class MESSENGERdata:
     query
         A SQL-style list of comparisons.
         
-    The data in the object created is extracted from the database tables using
+    load_spectra
+        Load the individual spectra as well as the radiance data.
+        Default = False
+        
+    The data in the object created is extracted from the database tables usingu
     the query:
     
     ::
@@ -75,7 +77,7 @@ class MESSENGERdata:
     model_strength
         If *N* models have been run, this is a dictionary in the form
         `{'model0':strength0, ..., 'modelN':strengthN}` containing modeled
-        source rates in units of :math:`10^{26}` atoms/s.
+        source rates in units of :math:`10^{23}` atoms/s.
         
     **Examples**
     
@@ -153,7 +155,7 @@ class MESSENGERdata:
     
         >>> inputs = Input('Ca.spot.Maxwellian.input')
         >>> CaData.model(inputs, 1e5, label='Model 1')
-        >>> inputs..speeddist.temperature /= 2.  # Run model with different temperature
+        >>> inputs.speeddist.temperature /= 2.  # Run model with different temperature
         >>> CaData.model(inputs, 1e5, label='Model 2')
         
     4. Plotting data
@@ -171,16 +173,15 @@ class MESSENGERdata:
 
     
     """
-    def __init__(self, species=None, comparisons=None):
+    def __init__(self, species=None, comparisons=None, load_spectra=False):
         allspecies = ['Na', 'Ca', 'Mg']
         self.species = None
         self.frame = None
         self.query = None
         self.data = None
         self.taa = None
-        self.inputs = None
-        self.model_info = None
-        
+        self.app = None
+        self.model_info = {}
         if species is None:
             pass
         elif species not in allspecies:
@@ -211,10 +212,31 @@ class MESSENGERdata:
                 self.species = species
                 self.frame = data.frame[0]
                 self.query = comparisons
-                data.drop(['species', 'frame'], inplace=True, axis=1)
-                self.data = data
-                self.data.set_index('unum', inplace=True)
                 self.taa = np.median(data.taa)
+                
+                data.drop(['species', 'frame'], inplace=True, axis=1)
+                data.loc[data.alttan < 0, 'alttan'] = 1e10
+                data.set_index('unum', inplace=True)
+                
+                if load_spectra:
+                    specquery = f'''SELECT *
+                                    FROM {species}spectra
+                                    WHERE snum in {data.index.to_list()}'''
+                    specquery = specquery.replace('[', '(').replace(']', ')')
+                    with database_connect() as con:
+                        spectra = pd.read_sql(specquery, con)
+                        
+                    spectra.set_index('snum', inplace=True)
+                    for r, row in spectra.iterrows():
+                        spectra.loc[r, 'wavelength'] = np.array(row.wavelength)
+                        spectra.loc[r, 'raw'] = np.array(row.raw)
+                        spectra.loc[r, 'solarfit'] = np.array(row.solarfit)
+                        spectra.loc[r, 'dark'] = np.array(row.dark)
+                        spectra.loc[r, 'calibrated'] = np.array(row.calibrated)
+                    self.data = data.merge(spectra, left_index=True,
+                                           right_index=True)
+                else:
+                    self.data = data
             else:
                 print(query)
                 print('No data found')
@@ -235,10 +257,7 @@ class MESSENGERdata:
         return result
 
     def __len__(self):
-        try:
-            return len(self.data)
-        except Exception:
-            return 0
+        return len(self.data)
 
     def __getitem__(self, q_):
         if isinstance(q_, int):
@@ -257,7 +276,6 @@ class MESSENGERdata:
         new.taa = self.taa
         new.data = self.data.iloc[q].copy()
         new.model_info = self.model_info
-        new.inputs = self.inputs
 
         return new
 
@@ -276,20 +294,20 @@ class MESSENGERdata:
 
         More frames could be added if necessary.
         If Frame is not specified, flips between MSO and Model."""
-        if (frame is None) and (self.frame == 'MSO'):
+        if (frame is None) and (self.frame.upper() == 'MSO'):
             frame = 'Model'
-        elif (frame is None) and (self.frame == 'Model'):
+        elif (frame is None) and (self.frame.upper() == 'MODEL'):
             frame = 'MSO'
         else:
             pass
 
-        allframes = ['Model', 'MSO']
-        if frame not in allframes:
+        allframes = ['MODEL', 'MSO']
+        if frame.upper() not in allframes:
             print('{} is not a valid frame.'.format(frame))
             return None
         elif frame == self.frame:
             pass
-        elif (self.frame == 'MSO') and (frame == 'Model'):
+        elif (self.frame.upper() == 'MSO') and (frame.upper() == 'MODEL'):
             # Convert from MSO to Model
             self.data.x, self.data.y = self.data.y.copy(), -self.data.x.copy()
             self.data.xbore, self.data.ybore = (self.data.ybore.copy(),
@@ -297,7 +315,7 @@ class MESSENGERdata:
             self.data.xtan, self.data.ytan = (self.data.ytan.copy(),
                                               -self.data.xtan.copy())
             self.frame = 'Model'
-        elif (self.frame == 'Model') and (frame == 'MSO'):
+        elif (self.frame.upper() == 'MODEL') and (frame.upper() == 'MODEL'):
             self.data.x, self.data.y = -self.data.y.copy(), self.data.x.copy()
             self.data.xbore, self.data.ybore = (-self.data.ybore.copy(),
                                                 self.data.xbore.copy())
@@ -307,9 +325,10 @@ class MESSENGERdata:
         else:
             assert 0, 'You somehow picked a bad combination.'
 
-    def model(self, inputs_, npackets, quantity='radiance',
-              fit_method='chisqmin', dphi=3*u.deg, overwrite=False,
-              masking=None, filenames=None, label=None):
+    def model(self, start_from, npackets, quantity='radiance',
+              fit_method='chisq', dphi=3*u.deg, overwrite=False,
+              masking=None, filenames=None, label=None,
+              fit_to_data=False, packs_per_it=None):
         """Run the nexoclom model with specified inputs and fit to the data.
         
         ** Parameters**
@@ -321,7 +340,7 @@ class MESSENGERdata:
             
         fit_method
             Allows user to specify the quantity to be minimized when fitting
-            the model to the data. Default = chisqmin (chi-squared
+            the model to the data. Default = chisq (chi-squared
             minimization). Other option is 'difference' which minimizes the
             sum of the differences of the data and the model ignoring the
             data uncertainty.
@@ -347,15 +366,21 @@ class MESSENGERdata:
               of the data in an initial fit to the data.
         """
 
-        if isinstance(inputs_, str):
-            inputs = Input(inputs_)
-        elif hasattr(inputs_, 'line_of_sight'):
-            inputs = inputs_
+        runmodel = True
+        if isinstance(start_from, str):
+            inputs = Input(start_from)
+            output = None
+        elif isinstance(start_from, Input):
+            inputs = copy.deepcopy(start_from)
+            output = None
+        elif isinstance(start_from, Output):
+            inputs = copy.deepcopy(start_from.inputs)
+            output = start_from
+            runmodel = False
         else:
             raise InputError('MESSENGERdata.model', 'Problem with the inputs.')
 
         # TAA needs to match the data
-        oldtaa = inputs.geometry.taa
         if len(self.data.orbit.unique()) == 1:
             inputs.geometry.taa = np.median(self.data.taa)*u.rad
         elif self.data.taa.max()-self.data.taa.min() < 3*np.pi/180.:
@@ -371,27 +396,35 @@ class MESSENGERdata:
             pass
 
         # Run the model
-        inputs.run(npackets, overwrite=overwrite)
+        self.set_frame('Model')
+        if runmodel:
+            inputs.run(npackets, overwrite=overwrite, packs_per_it=packs_per_it)
+            model_result = LOSResult(inputs, self, quantity, dphi=dphi,
+                                     filenames=filenames, overwrite=overwrite,
+                                     masking=masking, fit_to_data=fit_to_data)
+        else:
+            model_result = LOSResult(output, self, quantity, dphi=dphi,
+                                     filenames=filenames, overwrite=overwrite,
+                                     fit_method=fit_method, masking=masking,
+                                     fit_to_data=fit_to_data)
 
         # Simulate the data
-        if self.inputs is None:
-            self.inputs = [inputs]
-        else:
-            self.inputs.append(inputs)
-
-        self.set_frame('Model')
-        model_result = inputs.line_of_sight(self.data, quantity,
-                                            dphi=dphi, filenames=filenames,
-                                            overwrite=overwrite)
-        
         # modkey is the number for this model
-        modkey = f'model{len(self.inputs)-1:00d}'
-        packkey = f'packets{len(self.inputs)-1:00d}'
-        maskkey = f'mask{len(self.inputs)-1:00d}'
+        modnum = len(self.model_info)
+        modkey = f'model{modnum:00d}'
+        npackkey = f'npackets{modnum:00d}'
+        maskkey = f'mask{modnum:00d}'
         self.data[modkey] = model_result.radiance/1e3 # Convert to kR
-        self.data[packkey] = model_result.packets
+        self.data[npackkey] = model_result.ninview
+
+        strength, goodness_of_fit, mask = fit_model(self.data.radiance.values,
+                                                    self.data[modkey].values,
+                                                    self.data.sigma.values,
+                                                    fit_method=fit_method,
+                                                    masking=masking,
+                                                    altitude=self.data.alttan)
+        strength *= u.def_unit('10**23 atoms/s', 1e23/u.s)
         
-        strength, goodness_of_fit, mask = self.fit_model(modkey, fit_method, masking)
         self.data[modkey] = self.data[modkey]*strength.value
         self.data[maskkey] = mask
 
@@ -400,161 +433,33 @@ class MESSENGERdata:
         else:
             pass
 
-        model_info = {'fit_method': fit_method,
+        model_info = {'inputs': inputs,
+                      'fit_method': fit_method,
                       'goodness-of-fit': goodness_of_fit,
                       'strength': strength,
-                      'label': label}
-        if self.model_info is None:
-            self.model_info = {modkey: model_info}
-        else:
-            self.model_info[modkey] = model_info
-        
+                      'label': label,
+                      'filenames': model_result.filenames,
+                      'fitted': model_result.fitted,
+                      'savefile': model_result.savefiles}
+        self.model_info[modkey] = model_info
         print(f'Model strength for {label} = {strength}')
 
-        # Put the old TAA back in.
-        inputs.geometry.taa = oldtaa
-    
-    def fit_model(self, modkey, fit_method, masking):
-        def chisq(x):
-            return np.sum((self.data[mask]['radiance'] -
-                           x * self.data[mask][modkey])**2 /
-                          self.data[mask]['sigma']**2)/(sum(mask) - 1)
-
-        def difference(x):
-            return np.sum(np.abs(self.data[mask]['radiance'] -
-                                 x * self.data[mask][modkey]))
-
-        mask = np.array([True for _ in self.data.radiance])
-        sigmalimit = None
-        if masking is not None:
-            for masktype in masking.split(';'):
-                masktype = masktype.strip().lower()
-                if masktype.startswith('middle'):
-                    perinterval = float(masktype[6:])
-                    # Estimate model strength (source rate) by fitting middle %
-                    interval = PercentileInterval(perinterval)
-                    lim = interval.get_limits(self.data.radiance)
-                    mask = (mask &
-                            (self.data.radiance >= lim[0]) &
-                            (self.data.radiance <= lim[1]))
-                elif masktype.startswith('minalt'):
-                    minalt = float(masktype[6:])
-                    mask = mask & (self.data.alttan >= minalt)
-                elif masktype.startswith('minsnr'):
-                    minSNR = float(masktype[6:])
-                    snr = self.data.radiance / self.data.sigma
-                    mask = mask & (snr > minSNR)
-                elif masktype.startswith('siglimit'):
-                    sigmalimit = masktype
-                else:
-                    raise InputError('MESSENGERdata.fit_model',
-                                     f'masking = {masktype} not defined.')
-        else:
-            pass
-        
-        strunit = u.def_unit('10**26 atoms/s', 1e26/u.s)
-        available_fitfunctions = ['chisqmin', 'diffmin']
-        if fit_method.lower() in available_fitfunctions:
-            model_strength = minimize_scalar(fit_method.lower())
-            if sigmalimit is not None:
-                siglimit = float(sigmalimit[8:])
-                diff = ((self.data['radiance']-
-                         model_strength.x*self.data[modkey]) /
-                        self.data['sigma'])
-                mask = mask & (diff < siglimit*self.data['sigma'])
-                model_strength = minimize_scalar(fit_method.lower())
+    def plot(self, filename=None, plot_method='plotly', show=False):
+        if plot_method == 'plotly':
+            app = plot_plotly(self)
+            if show:
+                app.run_server(debug=False)
             else:
                 pass
+            self.app = app
+        elif plot_method == 'bokeh':
+            plot_bokeh(self, filename, show)
         else:
-            raise InputError('MESSENGERdata.fit_model',
-                             f'fit_method = {fit_method} not defined.')
-        
-        return model_strength.x * strunit, model_strength.fun, mask
-
-    def plot(self, filename=None, show=True, **kwargs):
-        if filename is not None:
-            if not filename.endswith('.html'):
-                filename += '.html'
-            else:
-                pass
-            bkp.output_file(filename)
-        else:
-            pass
-
-        # Format the date correction
-        self.data['utcstr'] = self.data['utc'].apply(
-            lambda x: x.isoformat()[0:19])
-
-        # Put the dataframe in a useable form
-        self.data['lower'] = self.data.radiance - self.data.sigma
-        self.data['upper'] = self.data.radiance + self.data.sigma
-        source = bkp.ColumnDataSource(self.data)
-
-        # Make the figure
-        fig = bkp.figure(plot_width=1200, plot_height=800,
-                         x_axis_type='datetime',
-                         title=f'{self.species}, {self.query}',
-                         x_axis_label='UTC',
-                         y_axis_label='Radiance (kR)',
-                         tools=['pan', 'box_zoom', 'reset', 'save'])
-
-        # plot the data
-        dplot = fig.circle(x='utc', y='radiance', size=7, color='black',
-                           legend_label='Data', hover_color='white',
-                           source=source)
-        
-        # Add error bars
-        fig.add_layout(Whisker(source=source, base='utc', upper='upper',
-                               lower='lower'))
-
-        # tool tips
-        tips = [('index', '$index'),
-                ('UTC', '@utcstr'),
-                ('Radiance', '@radiance{0.2f} kR')]
-        if self.model_info is not None:
-            for modkey, info in self.model_info.items():
-                tips.append((info['label'], f'@{info["strength"]}{{0.2f}} kR'))
-        datahover = HoverTool(tooltips=tips,
-                              renderers=[dplot])
-        fig.add_tools(datahover)
-
-        # Plot the model
-        col = (c for c in Set1[9])
-        if self.model_info is not None:
-            for modkey, info in self.model_info.items():
-                try:
-                    c = next(col)
-                except StopIteration:
-                    col = (c for c in Set1[9])
-                    c = next(col)
-
-                f0 = fig.line(x='utc', y=modkey, source=source,
-                               legend_label=info['label'], color=c)
-                f1 = fig.circle(x='utc', y=modkey, size=7, source=source,
-                               legend_label=info['label'], color=c)
-                datahover.renderers.append(f0)
-                datahover.renderers.append(f1)
-
-        # Labels, etc.
-        fig.title.align = 'center'
-        fig.title.text_font_size = '16pt'
-        fig.axis.axis_label_text_font_size = '16pt'
-        fig.axis.major_label_text_font_size = '16pt'
-        fig.legend.label_text_font_size = '16pt'
-        fig.legend.click_policy = 'hide'
-
-        if filename is not None:
-            bkp.output_file(filename)
-            export_png(fig, filename=filename.replace('.html', '.png'))
-            bkp.save(fig)
+            print('Not a valid plotting method')
             
-        else:
-            pass
-
-        if show:
-            bkp.show(fig)
-
-        return fig
+    def plot_fitted_model(self, filestart='fitted', show=False,
+                          make_frames=False, smooth=False):
+        plot_fitted(self, filestart, show, make_frames, smooth=smooth)
 
     def export(self, filename, columns=('utc', 'radiance')):
         """Export data and models to a file.
@@ -577,7 +482,7 @@ class MESSENGERdata:
         
         """
         columns_ = list(columns)
-        if self.model_info is not None:
+        if len(self.model_info) != 0:
             columns_.extend(self.model_info.keys())
         else:
             pass
@@ -616,5 +521,3 @@ class MESSENGERdata:
             subset.to_latex(filename)
         else:
             print('Valid output formats = csv, pkl, html, tex')
-
-
